/*
* This file is part of fwmon
* Copyright (c) 2001,2002 Gianni Tedesco
*
*    fwmon is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    fwmon is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with fwmon; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

static const char rcsid[]=
	"@(#) $Id: main.c,v 1.22 2002/02/12 18:58:21 gianni Exp $";

#include "g_globals.h"
#include "main.h"
#include "netlink.h"
#include "sniffer.h"
#include "print.h"
#include "slog.h"

#ifdef _HAVE_LIBPCAP
#include "libpcap.h"
#endif

#include <time.h>
#include <sys/utsname.h>
#include <signal.h>

struct globals_t globals;

cap_open capture_open;
cap_close capture_close;
cap_mainloop capture_mainloop;

/* Decide what to do with a packet */
void dump_packet ( p_packet p )
{
	/* Perform basic demultiplexing */
	p->trans = (void *) ((long)p->ip) + (p->ip->ihl*4);
	p->app = NULL;

	if ( (flag(FLAG_STDOUT) && !flag(FLAG_DAEMON) )
		|| flag(FLAG_LOGFILE) ) {
		print_dump(p);
	}

	if ( flag(FLAG_SYSLOG) ) {
		slog_dump(p);
	}

#ifdef _HAVE_LIBPCAP
	if ( flag(FLAG_LIBPCAP) ) {
		lpcap_dump(p);
	}
#endif
}

/* Display usage information */
void usage(char *apppath)
{
	if (apppath==NULL) apppath="";

	print_err("Usage:" CRLF);
	print_err("  %s [OPTIONS]" CRLF CRLF, apppath);

	/* Output */
	print_err("  -t <filename>\tName of file to output tcpdump logs" CRLF);
	print_err("  -l <filename>\tName of file to log to" CRLF);
	print_err("  -q \t\tLog as SQL" CRLF);
	print_err("  -s \t\tSilent, no output to screen" CRLF);
	print_err("  -d \t\tDaemon mode, silent, and backgrounded" CRLF);
	print_err("  -y \t\tLog machine parsable output to syslog" CRLF);

	/* Formatting */
	print_err("  -n <length>\tLine length for hex/ascii output" CRLF);
	print_err("  -a \t\tShow the date as well as time on packets" CRLF);
	print_err("  -m \t\tDo not dump data" CRLF);
	print_err("  -x \t\tDo not display hex output" CRLF);
	print_err("  -i \t\tDo not display ascii output" CRLF);

	/* Security */
	print_err("  -u <uid>\tUser ID to drop to" CRLF);
	print_err("  -g <gid>\tGroup ID to drop to" CRLF);
	print_err("  -c <chroot>\tDirectory to chroot jail inside" CRLF);

	/* Help */
	print_err("  -v \t\tJust print version" CRLF);
	print_err("  -h \t\tDisplay this help" CRLF CRLF);
	print_err("  Report bugs to: Gianni Tedesco <gianni@ecsc.co.uk>" CRLF CRLF);

	if ( err(ERROR_SUGID) ){
		print_err("%s: " ERR_SUGID CRLF, apppath);
	}
	if ( err(ERROR_FMATCH) ){
		print_err("%s: " ERR_FMATCH CRLF, apppath);
	}
	if ( err(ERROR_NODATA) ){
		print_err("%s: " ERR_NODATA CRLF, apppath);
	}
	if ( err(ERROR_ZEROLINE) ){
		print_err("%s: " ERR_ZEROLINE CRLF, apppath);
	}	
	if ( err(ERROR_LIBPCAP) ){
		print_err("%s: " ERR_LIBPCAP CRLF, apppath);
	}
	if ( err(ERROR_UNAME) ){
		print_err("%s: " ERR_UNAME CRLF, apppath);
	}
	if ( err(ERROR_NOOUT) ){
		print_err("%s: " ERR_NOOUT CRLF, apppath);
	}
	
}

/* Parse the command line for options */
int parse_command_line( int argc, char **argv )
{
	int i, ret=1;
	struct utsname un;

	while ( (i=getopt(argc,argv, "t:l:sdhvu:g:c:axmin:yq" ))!=-1){
    	switch (i)
    	{
		case 'q':
			flag_set(FLAG_SQL);
			if ( !uname(&un) ) {
				globals.uname=strdup((char *)&un.nodename);
			}else{
				err_set(ERROR_UNAME);
				ret = 0;
			}
			continue;
		case 'y':
			flag_set(FLAG_SYSLOG);
			continue;
    		case 'n':
    			globals.length = atoi(optarg);
    			continue;
    		case 'i':
    			flag_set(FLAG_NOASCII);
    			continue;
    		case 'x':
    			flag_set(FLAG_NOHEX);
    			continue;
    		case 'm':
    			flag_set(FLAG_NODATA);
    			continue;
    		case 'a':
    			flag_set(FLAG_SHOWDATE);
    			continue;
    		case 'c':
    			flag_set(FLAG_CHROOT);
    			globals.chroot = optarg;
    			continue;
    		case 'u':
    			flag_set(FLAG_SUID);
    			globals.uid = atoi(optarg);
    			continue;
    		case 'g':
    			flag_set(FLAG_SGID);
    			globals.gid = atoi(optarg);
    			continue;
    		case 't':
#ifdef _HAVE_LIBPCAP
    			flag_set(FLAG_LIBPCAP);
    			globals.pcapfile=optarg;
#else
			err_set(ERROR_LIBPCAP);
			ret=0;
#endif
			continue;
    		case 'l':
    			flag_set(FLAG_LOGFILE);
    			globals.logfile=optarg;
    			continue;
    		case 's':
 			flag_unset(FLAG_STDOUT);
    			continue;
    		case 'd':
    			flag_unset(FLAG_STDOUT);
    			flag_set(FLAG_DAEMON);
    			continue;
    		case 'v':
    			print_err("%s: " VERSION_STRING CRLF CRLF, argv[0]);
    			exit(0);
    		case 'h':
    		case '?':
    		default:
    			return 0;
    	}
	}

	
	if ( (flag(FLAG_SUID) || flag(FLAG_SGID)) && 
			!(flag(FLAG_SUID) && flag(FLAG_SGID)) ) {
		err_set(ERROR_SUGID);
		ret = 0;
	}

	if ( flag(FLAG_LIBPCAP) && flag(FLAG_LOGFILE) ) {
		if ( !strcmp(globals.logfile, globals.pcapfile) ) {
			err_set(ERROR_FMATCH);
			ret = 0;
		}
	}
	
	/* AUDIT: length must be 1 or more,
	 * not 0 or more, otherwise hex dump
	 * loops infinately */
	if ( globals.length <= 0 ) {
		err_set(ERROR_ZEROLINE);
		ret = 0;
	}

	if ( flag(FLAG_NOASCII) && flag(FLAG_NOHEX) ) {
		err_set(ERROR_NODATA);
		ret = 0;
	}

	if ( !flag(FLAG_STDOUT) && !flag(FLAG_LOGFILE) &&
		!flag(FLAG_LIBPCAP) && !flag(FLAG_SYSLOG) ) {
		err_set(ERROR_NOOUT);
		ret=0;
	}
	
	return ret;
}

/* DAEMONIZE - drops root privs, and chroots etc.. */
int daemonize(void)
{
	if ( flag(FLAG_CHROOT) ) {
		if ( chdir(globals.chroot) == -1 ) {
			perror("chdir()");
			return 0;
		}
		if ( chroot(".") == -1 ) {
			perror("chroot()");
			return 0;
		}
	}

	if ( flag(FLAG_SGID) ) {
		if (setgroups(1,&globals.gid) == -1) {
			perror("setgroups()");
			return 0;
		}
		
		/* AUDIT: Check we actually dropped GID! */
		if ( setgid(globals.gid) == -1 ) {
			perror("setgid()");
			return 0;
		}
	}

	if ( flag(FLAG_SUID) ){
		/* AUDIT: Check we actually dropped UID! */
		if ( setuid(globals.uid) == -1 ) {
			perror("setuid()");
			return 0;
		}
	}

	/* background our process */
	if ( flag(FLAG_DAEMON) ) {
		switch(fork())
		{
			case 0: break;
			case -1:
				perror("fork()");
				return 0;
			default: _exit(0);
		}

		if (setsid() < 0) {
			perror("setsid()");
			return 0;
		}
		
		switch(fork())
		{
			case 0: break;
			case -1:
				perror("fork() again");
				return 0;
			default: _exit(0);
		}
	}

	return 1;
}


/* APP ENTRY POINT */
int main ( int argc, char **argv )
{
	int fd=0, fds=sysconf(_SC_OPEN_MAX);

	/* zero global data */
	memset(&globals, 0, sizeof(globals));

	/* Setup capture */
#ifndef SNIFFER_NOT_NETLINK
	capture_mainloop=netlink_mainloop;
	capture_open=netlink_open;
	capture_close=netlink_close;
#else
	capture_mainloop=sniffer_mainloop;
	capture_open=sniffer_open;
	capture_close=sniffer_close;
#endif	
	/* Set default flags */
	flag_set(FLAG_STDOUT);
	globals.length = LINE_LENGTH;
	
	/* Set timezone */
	tzset();

	if ( !parse_command_line(argc, argv) ) {
		usage(argv[0]);
		exit(1);
	}

#ifdef _DEBUG
	if ( flag(FLAG_LIBPCAP) )
		print_err("TCPDUMP file: %s" CRLF, globals.pcapfile);
	if ( flag(FLAG_LOGFILE) )
		print_err("LOGFILE file: %s" CRLF, globals.logfile);
	if ( flag(FLAG_SYSLOG) ) print_err("SYSLOG" CRLF);
	if ( flag(FLAG_SQL) ) print_err("POSTGRESQL: %s" CRLF, globals.sqlfile);
	if ( flag(FLAG_CHROOT) ) print_err("CHROOT: %s" CRLF, globals.chroot);
	if ( flag(FLAG_SUID) ) print_err("SUID: %i" CRLF, globals.uid);
	if ( flag(FLAG_SGID) ) print_err("SGID: %i" CRLF, globals.gid);
	if ( flag(FLAG_DAEMON) )  print_err("DAEMON MODE" CRLF);
	if ( !flag(FLAG_STDOUT) )  print_err("SILENT MODE" CRLF);
#endif

	/* Need this _before_ we open netlink/output files ;) */
	if ( flag(FLAG_DAEMON) ) {
		while (fd<fds) close(fd++);
	}

	if ( !capture_open() ) {
		fprintf(stderr, ERR_OPEN_CAP CRLF);
		exit(1);
	}

	if ( !daemonize() ) exit(1);
	
	/* Initialise various output mechanisms
	 * this opens file descriptors */
	if (!print_init()) {
		print_err("print_init()" CRLF);
		exit(1);
	}

	if (!slog_init()) {
		print_err("slog_init()" CRLF);
		exit(1);
	}

#ifdef _HAVE_LIBPCAP
	if (!lpcap_init()) {
		print_err("lpcap_init()" CRLF);
		print_close();
		exit(1);
	}
#endif
	capture_mainloop();

#ifdef _HAVE_LIBPCAP
	lpcap_close();
#endif
	print_close();
	capture_close();
	exit(1);
}
